#!/bin/python3
# Copyright (C) 2022, Dr. Danny Bickson and Dr. Amir Alush.
import sys
import os
import cv2
from ctypes import *
import pandas as pd
pd.set_option('display.max_colwidth', None)
import numpy as np
import traceback
import platform

__version__ = "0.65"
CONTACT_EMAIL="info@databasevisual.com"

if platform.system() == "Darwin":
	SO_SUFFIX=".dylib"
else:
	SO_SUFFIX=".so"

LOCAL_DIR=os.path.dirname(os.path.abspath(__file__))
so_file = os.path.join(LOCAL_DIR, 'libfastdup_shared' + SO_SUFFIX)
if not os.path.exists(so_file):
    print("Failed to find shared object", so_file);
    print("Current init file is on", __file__);
    sys.exit(1)
dll = CDLL(so_file)

model_path_full=os.path.join(LOCAL_DIR, 'MobileNetV3Small_feat_opt.onnx')
if not os.path.exists(model_path_full):
    print("Failed to find onnx file", model_path_full);
    print("Current init file is on", __file__);
    sys.exit(1)


"""
This is a wrapper function to call the fastdup c++ code
"""

def run(input_dir='',             
        work_dir='.', 
        test_dir='',
        compute='cpu',    
        verbose=False,     
        num_threads=-1,     
        num_images=0,        
        turi_param='nnmodel=0',  
        distance='cosine',     #distance metric for the nearest neighbor model. Default "cosine".
        threshold=0.9,         #threshold for finding simiar images. Default is 0.85 (values 0->1)
        lower_threshold=0.05,   #lower threshold for finding simiar images. Default is 0.05 (values 0->1)
        output_similarity='similarity.csv', #output similarity csv file location
        model_path=model_path_full,
        #ONNX model path
        license='CC-BY-NC-ND-4.0',            #license string
        version=False,          #show version and exit      
        nearest_neighbors_k=2, 
        d=576,
        run_mode=0,
        nn_provider='faiss',
        min_offset=0,
        max_offset=0, 
        faiss_mode="HNSW32",
        faiss_param="",
        bounding_box=""): 

    '''
    Run fastdup tool for find duplicate and near duplicate images in a corpus of images. 
    The only mandatory argument is image_dir. Given an image directory it will compare all pairs of images and store the most similar ones in the output file output_similarity.

    Parameters:
        input_dir (str): Location of the images. Either a folder, of a file containing absolute filenames each on its own row. We support jpg, jpeg, tiff, tif, giff, png. In addition we support tar, tar.gz, tgz and zip files containing images.   Locoation could be a local folder, or s3 bucket or minio remote file (starting with minio://).

        work_dir (str): Optional path for storing intermediate files and results. Default is "."

        test_dir (str): Optional path for test data. When given similarity of train and test images is compared (vs. train/train or test/test which are not performed).

        compute (str): Compute type [cpu|gpu] default is cpu.

        verbose (boolean): Verbosity. Default is False.

        num_threads (int): Number of threads. Default is -1 to be auto configured by the number of cores.

        num_images (unsigned long long): Number of images to run on. Default is 0 which means run on all the images in the image_dir folder.

        turi_param (str): Optional turi parameters seperated by command. 
            nnmodel=xx, Nearest Neighbor model for clustering the features together. Supported options are 0 = brute_force (exact), 1 = ball_tree and 2 = lsh (both approximate).
            ccthreshold=xx, Threshold for running conected components to find clusters of similar images. Allowed values 0->1.
            store_int=0|1 store the similarity as string filenames or string index of the file id (to save space)
            Example run: turi_param='nnmodel=0,ccthreshold=0.99'

        distance (str): Distance metric for the Nearest Neighbors algorithm. Default is cosine. Other distances are euclidean, squared_euclidean, manhattan.

        threshold (float): Similarity measure in the range 0->1, where 1 is totally identical, 0.98 and above is almost identical, and 0.85 and above is very similar. Default is 0.85 which means that only image pairs with similarity larger than 0.85 are stored.

        lower_threshold (float): Similarity measure to outline images that are far away (outliers) vs. the total distribution. Default value is 0.05 (means 5% out of the total similarities computed).

        model_path(str): Optional location of ONNX model file, should not be used.

        version(bool): Print out the version number. This function takes no argument.

        nearest_neighbors_k (int): For each image, how many similar images to look for. Default is 2.

        d (int): Length of the feature vector. Default is 576. Change this parameter only when providing precomputed features instead of images.

        run_mode (int): This software can run for either feature vector extraction and similarity measurement (0), or just feature vector extraction (1), or just similarity measure computation (2).

        nn_provider (string): Provider of the nearest neighbor algorithm, allowed values are turi|faiss.

        min_offset (unsigned long long): Optional min offset to start iterating on the full file list. Default is -1.

        max_offset (unsigned long long): Optional max offset to start iterating on the full file list. Default is -1.

        faiss_mode (str): When nn_provider='faiss' selects the faiss mode. Default option is HNSW32.

        faiss_param (str): When nn_provider='faiss' selects assigns optional faiss parameters. For example efSearch=175. Multiple params are supported for example 'efSearch=175,nprobes=200'

        bounding_box (str): Optional bouding box to crop images, given as bounding_box='rows=xx,cols=xx,height=xx,width=xx'.

    Returns:
        Status code 0 = success, 1 = error.

    '''


    if (version):
        print("FastDup Software, (C) copyright 2022 Dr. Amir Alush and Dr. Danny Bickson.");
        print("This software is free for non-commercial and academic usage under the Creative Common Attribution-NonCommercial-NoDerivatives 4.0 "
                "International license. Please reach out to %s for licensing options.", CONTACT_EMAIL);
        return 0


    if (input_dir.strip() == '' and run_mode != 2):
        print("Found an empty input directory, please point to the directory where you are images are found");
        return 0


    #Calling the C++ side
    dll.do_main.restype = c_int
    dll.do_main.argtypes = [c_char_p, 
            c_char_p,
            c_char_p,
            c_char_p,
            c_bool,
            c_int,
            c_ulonglong, 
            c_char_p,
            c_char_p,
            c_float,
            c_float,
            c_char_p,
            c_char_p,
            c_bool,
            c_int,
            c_int,
            c_int,
            c_char_p,
            c_ulonglong,
            c_ulonglong,
            c_char_p,
            c_char_p, 
            c_char_p]


    dll.do_main(bytes(input_dir, 'utf-8'),
            bytes(work_dir, 'utf-8'),
            bytes(test_dir.strip(), 'utf-8'),
            bytes(compute, 'utf-8'), 
            verbose, 
            num_threads, 
            num_images,
            bytes(turi_param, 'utf-8'),
            bytes(distance, 'utf-8'),
            threshold, 
            lower_threshold,
            bytes(license, 'utf-8'),
            bytes(model_path, 'utf-8'), 
            version,
            nearest_neighbors_k,
            d,
            run_mode,
            bytes(nn_provider, 'utf-8'),
            min_offset,
            max_offset,
            bytes(faiss_mode, 'utf-8'),
            bytes(faiss_param, 'utf-8'), 
            bytes(bounding_box, 'utf-8'))




def load_binary_feature(filename, d=576):
    '''
    Python function for loading the stored binary features written by fastdup and their matching filenames and analyzing them in Python.

    Parmaeters:
        filename(str):The binary feature file location
        d(int): Feature vector length

    Returns:
        A list of with all image file names of length X.
        An np matrixy of shape rows x d cols (default d is 576). Each row conform to feature vector os a single image.

    Example:
        import fastdup
        file_list, mat_features = fastdup.load_binary('features.dat')

    '''
	
    if not os.path.exists(filename) or not os.path.exists(filename + '.csv'):
        print("Error: failed to find the binary feature file:", filename, ' and the filenames csv file:', filename + '.csv');
        return None;
    assert(d > 0), "Feature vector length d has to be larger than zero"

    with open(filename, 'rb') as f:
        data = np.fromfile(f, dtype='<f')

    df = pd.read_csv(filename + '.csv')['filename'].values;
    num_images = len(df);
    print('Read a total of ', num_images, 'images');

    data = np.reshape(data, (num_images, d));
    assert data.shape[1] == d
    return list(df), data



def save_binary_feature(save_path, filenames, np_array):
    '''
    Function for saving data to be used by fastdup. Given a list of images and their matching feature vectors in a numpy array,
    function saves data in a format readable by fastdup. This saves the image extraction step, to be used with run_mode=1 namely perform
    nearest neighbor model on the feature vectors.

    Parmaeters:
        save_path(str): Working folder to save the files to
        filenames(list): A list of file location of the images (absolute paths) of length n images
        np_array(np.array): Numpy array of size n x d. Each row is a feature vector of one file.

    Returns:
        0 in case of sucess, otherwise 1

    '''

    assert isinstance(save_path, str)  and save_path.strip() != "", "Save path should be a non empty string"
    assert isinstance(filenames, list), "filenames should be a list of image files"
    assert len(filenames), "filenames should be a non empty list"
    assert isinstance(filenames[0], str), 'filenames should contain strings with the image absolute paths'
    assert isinstance(np_array, np.ndarray),  "np_array should be a numpy array"
    assert np_array.dtype == 'float32', "np_array dtype must be float32. You can generate the array using the" \
                              "command: features = np.zeros((rows, cols), dtype='float32')"
    assert np_array.shape[0] == len(filenames), "np_array should contain rows matching to the filenames list"

    
    try:
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        
        df = pd.DataFrame({'filename': filenames})
        local_filename = os.path.join(save_path, 'features.dat')
        df.to_csv(local_filename + '.csv', index=False)
        bytes_array = np_array.tobytes()
        with open(local_filename, 'wb') as f:
            f.write(bytes_array)
        assert os.path.exists(local_filename), "Failed to save file " + local_filename

    except Exception as ex:
        print("Failed to save to " + save_path + " Exception: " + ex)
        return 1

    return 0


def get_label(filename, get_label_func):
    ret = filename
    try:
        ret += "<br>" + "Label: " + get_label_func(filename)
    except Exception as ex:
        ret += "<br>Failed to get label for " + filename + " with error " + ex
    return ret



def extract_filenames(row):
    impath1 = row[1]['from']
    impath2 = row[1]['to']
    dist = row[1]['distance']
    os.path.exists(impath1), "Failed to find image file " + impath1
    os.path.exists(impath2), "Failed to find image file " + impath2

    type1 = image.get_type(impath1)
    type2 = image.get_type(impath2)
    ptype = '{0}_{1}'.format(type1, type2)
    return impath1, impath2, dist, ptype
 

def create_duplicates_gallery(similarity_file, save_path, num_images=20, descending=True, 
                            lazy_load=False, get_label_func=None):
    '''

    Function to create and display a gallery of images computed by the similarity metrics

    Parameters:
        similarity_file (str): csv file with the computed similarities by the fastdup tool
        save_path (str): output folder location for the visuals
        num_images(int): Max number of images to display (default = 50). Be careful not to display too many images at once otherwise the notebook may go out of memory.
        descending (boolean): If False, print the similarities from the least similar to the most similar. Default is True.

        lazy_load (boolean): If False, write all images inside html file using base64 encoding. Otherwise use lazy loading in the html to load images when mouse curser is above the image (reduced html file size).

        get_label_func (callable): Optional parameter to allow adding more image information to the report like the image label. This is a function the user implements that gets the full file path and returns html string with the label or any other metadata desired.
    '''
    assert os.path.exists(similarity_file), "Failed to find similarity file " + similarity_file
    assert num_images >= 1, "Please select one or more images"
    if num_images > 1000 and not lazy_load:
        print("When plotting more than 1000 images, please run with lazy_load=True. Chrome and Safari support lazy loading of web images, otherwise the webpage gets too big")
    assert os.path.exists(save_path), "Failed to find save_path " + save_path
    if (get_label_func is not None):
        assert callable(get_label_func), "get_label_func has to be a collable function, given the filename returns the label of the file"

    from fastdup import image
    try:
        from tqdm import tqdm
    except:
        tqdm = (lambda x: x)

    images = []
    img_paths = []
    df = pd.read_csv(similarity_file)
    assert len(df), "Failed to read similarity file"

    subdf = df.head(num_images) if descending else df.tail(num_images)
    for row in tqdm(subdf.iterrows(), total=num_images):
        impath1, impath2, dist, ptype = extract_filenames(row)
        try:
            img, imgpath = image.create_triplet_img(impath1, impath2, ptype, dist, save_path)
        except Exception as ex:
            traceback.print_exc()
            print("Failed to generate viz for images", impath1, impath2, ex)
            img = None
            imgpath = None
        img_paths.append(imgpath)
    
    import fastdup.html_writer

    if not lazy_load:
        subdf.insert(0, 'Image', [image.imageformatter(x) for x in img_paths])
    else:
        img_paths2 = ["<img src=\"" + os.path.basename(x) + "\" loading=\"lazy\">" for x in img_paths]
        subdf.insert(0, 'Image', img_paths2)
    out_file = os.path.join(save_path, 'similarity.html')
    if get_label_func is not None and callable(get_label_func):
        subdf.insert(2, 'From', subdf['from'].apply(lambda x: get_label(x, get_label_func)))
        subdf.insert(3, 'To', subdf['to'].apply(lambda x: get_label(x, get_label_func)))
    else:
        subdf = subdf.rename(columns={'from':'From', 'to':'To'}, inplace=False)
    subdf = subdf.rename(columns={'distance':'Distance'}, inplace=False)

    fields = ['Image', 'Distance', 'From', 'To']



    fastdup.html_writer.write_to_html_file(subdf[fields], 'Fastdup tool - similarity report', out_file)
    assert os.path.exists(out_file), "Failed to generate out file " + out_file
    print("Stored similarity visual view in ", out_file)


def create_outliers_gallery(outliers_file, save_path, num_images=20, lazy_load=False, get_label_func=None):
    '''

    Function to create and display a gallery of images computed by the outliers metrics

    Parameters:
        outliers_file (str): csv file with the computed outliers by the fastdup tool
        save_path (str): output folder location for the visuals

        num_images(int): Max number of images to display (default = 50). Be careful not to display too many images at once otherwise the notebook may go out of memory.

        lazy_load (boolean): If False, write all images inside html file using base64 encoding. Otherwise use lazy loading in the html to load images when mouse curser is above the image (reduced html file size).

        get_label_func (callable): Optional parameter to allow adding more image information to the report like the image label. This is a function the user implements that gets the full file path and returns html string with the label or any other metadata desired.
     '''
    assert os.path.exists(outliers_file), "Failed to find similarity file " + similarity_file
    if num_images > 1000 and not lazy_load:
        print("When plotting more than 1000 images, please run with lazy_load=True. Chrome and Safari support lazy loading of web images, otherwise the webpage gets too big")
    assert os.path.exists(save_path), "Failed to find save_path " + save_path
    assert num_images >= 1, "Please select one or more images"
    if (get_label_func is not None):
        assert callable(get_label_func), "get_label_func has to be a collable function, given the filename returns the label of the file"

    from fastdup import image
    try:
        from tqdm import tqdm
    except:
        tqdm = (lambda x: x)

    img_paths = []
    df = pd.read_csv(outliers_file)
    assert len(df), "Failed to read outliers file"

    subdf = df.tail(num_images).sort_values(by='distance', ascending=True)
    for row in tqdm(subdf.iterrows(), total=num_images):
        impath1, impath2, dist, ptype = extract_filenames(row)
        try:
            img = cv2.imread(impath1)
            h, w, c = img.shape
            w1 = 320
            img = cv2.resize(img, (w1, w1))
            #consider saving second image as well!
            imgpath = os.path.join(save_path, os.path.basename(impath1))
            cv2.imwrite(imgpath, img)
            assert os.path.exists(imgpath), "Failed to save img to " + imgpath

        except Exception as ex:
            traceback.print_exc()
            print("Failed to generate viz for images", impath1, impath2, ex)
            imgpath = None
        img_paths.append(imgpath)

    import fastdup.html_writer
    if not lazy_load:
        subdf.insert(0, 'Image', [image.imageformatter(x) for x in img_paths])
    else:
        img_paths2 = ["<img src=\"" + os.path.join(save_path, os.path.basename(x)) + "\" loading=\"lazy\">" for x in img_paths]
        subdf.insert(0, 'Image', img_paths2)
    
    if get_label_func is not None and callable(get_label_func):
        subdf.insert(2, 'Path', subdf['from'].apply(lambda x: get_label(x, get_label_func)))
        subdf = subdf.rename(columns={'distance':'Distance'}, inplace=False)
    else:
        subdf = subdf.rename(columns={'from':'Path', 'distance':'Distance'}, inplace=False)
    
    out_file = os.path.join(save_path, 'outliers.html')
    fastdup.html_writer.write_to_html_file(subdf[['Image','Distance','Path']], 'Fastdup Tool - Outliers Report', out_file)
    assert os.path.exists(out_file), "Failed to generate out file " + out_file
    
    print("Stored outliers visual view in ", os.path.join(out_file))
